package cc.fascinated.bat.features.minecraft.command.minecraft;

import cc.fascinated.bat.command.BatCommand;
import cc.fascinated.bat.command.CommandInfo;
import cc.fascinated.bat.common.EmbedDescriptionBuilder;
import cc.fascinated.bat.common.EmbedUtils;
import cc.fascinated.bat.model.BatGuild;
import cc.fascinated.bat.model.BatUser;
import lombok.NonNull;
import net.dv8tion.jda.api.entities.Member;
import net.dv8tion.jda.api.entities.channel.middleman.MessageChannel;
import net.dv8tion.jda.api.interactions.commands.OptionMapping;
import net.dv8tion.jda.api.interactions.commands.OptionType;
import net.dv8tion.jda.api.interactions.commands.SlashCommandInteraction;
import net.dv8tion.jda.api.interactions.commands.build.OptionData;
import org.springframework.stereotype.Component;
import xyz.mcutils.McUtilsAPI;
import xyz.mcutils.models.cache.CachedBedrockMinecraftServer;
import xyz.mcutils.models.cache.CachedJavaMinecraftServer;
import xyz.mcutils.models.server.MinecraftServer;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@CommandInfo(
        name = "lookup-server",
        description = "Lookup a Minecraft server"
)
public class LookupServerSubCommand extends BatCommand {
    public LookupServerSubCommand() {
        super.addOptions(
                new OptionData(OptionType.STRING, "platform", "The platform of the server to lookup", true)
                        .addChoice("Java", "java")
                        .addChoice("Bedrock", "bedrock"),
                new OptionData(OptionType.STRING, "host", "The host/ip of the server to lookup", true)
        );
    }

    @Override
    public void execute(BatGuild guild, @NonNull BatUser user, @NonNull MessageChannel channel, Member member, @NonNull SlashCommandInteraction event) {
        OptionMapping platformOption = event.getOption("platform");
        OptionMapping hostOption = event.getOption("host");
        if (platformOption == null || hostOption == null) {
            return;
        }
        String platform = platformOption.getAsString();
        String host = hostOption.getAsString();

        MinecraftServer server;
        try {
            if (platform.equalsIgnoreCase("java")) {
                server = McUtilsAPI.getJavaServer(host);
            } else {
                server = McUtilsAPI.getBedrockServer(host);
            }
        } catch (Exception ex) {
            event.replyEmbeds(EmbedUtils.errorEmbed()
                    .setDescription("The server `%s` is invalid or offline".formatted(host))
                    .build()).queue();
            return;
        }

        int platformDefaultPort = platform.equalsIgnoreCase("java") ? 25565 : 19132;
        String hostname = server.getHostname() + (server.getPort() != platformDefaultPort ? ":" + server.getPort() : "");
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("Server Lookup [(Raw Data)](%s)".formatted(
                "https://api.mcutils.xyz/server/%s/%s".formatted(platform, hostname)
        ));
        description.appendLine("Host: `%s`".formatted(hostname), true);

        MinecraftServer.GeoLocation location = server.getLocation();
        if (server instanceof CachedJavaMinecraftServer javaServer) {
            description.appendLine("Version: `%s`".formatted(javaServer.getVersion().getName()), true);
            if (javaServer.getForgeData() != null) {
                description.appendLine("Forge Mods: `%s`".formatted(javaServer.getForgeData().getMods().length), true);
            }
            if (javaServer.isMojangBlocked()) {
                description.appendLine("Mojang Blocked: `Yes`", true);
            }
            if (javaServer.isEnforcesSecureChat()) {
                description.appendLine("Enforces Secure Chat: `Yes`", true);
            }
            if (javaServer.isPreventsChatReports()) {
                description.appendLine("Prevents Chat Reports: `Yes`", true);
            }
        }
        if (server instanceof CachedBedrockMinecraftServer bedrockServer) {
            description.appendLine("Version: `%s`".formatted(bedrockServer.getVersion().getName()), true);
        }
        if (location != null) {
            description.appendLine("Location: [%s](%s)".formatted(
                    (location.getCity() == null ? "" : location.getCity() + ", ") + location.getCountry(),
                    "https://www.google.com/maps/search/?api=1&query=%s,%s".formatted(location.getLatitude(), location.getLongitude())
            ), true);
        }

        event.replyEmbeds(EmbedUtils.successEmbed()
                .setDescription(description.build())
                .setThumbnail("https://api.mcutils.xyz/server/icon/%s".formatted(hostname))
                .setImage(server.getMotd().getPreview())
                .setFooter("Powered by mcutils.xyz")
                .build()).queue();
    }
}
